==> org-lsst-ccs-subsystem-image-handling-main-1.1.8-20241109.005552-4-sources.jar: org/lsst/ccs/subsystem/imagehandling/BusMessageHandler.java <==
package org.lsst.ccs.subsystem.imagehandling;

import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.lsst.ccs.Agent;
import org.lsst.ccs.bus.data.AgentCategory;
import org.lsst.ccs.bus.data.AgentInfo;
import org.lsst.ccs.bus.data.AgentInfo.AgentType;
import org.lsst.ccs.bus.data.ConfigurationInfo;
import org.lsst.ccs.bus.data.ConfigurationParameterInfo;
import org.lsst.ccs.bus.messages.BusMessage;
import org.lsst.ccs.bus.messages.StatusConfigurationInfo;
import org.lsst.ccs.bus.messages.StatusMessage;
import org.lsst.ccs.bus.messages.StatusSubsystemData;
import org.lsst.ccs.commons.annotations.LookupField;
import org.lsst.ccs.daq.ims.DAQException;
import org.lsst.ccs.daq.ims.ImageMetaData;
import org.lsst.ccs.framework.HasLifecycle;
import org.lsst.ccs.messaging.BusMessageFilterFactory;
import org.lsst.ccs.messaging.StatusMessageListener;
import org.lsst.ccs.subsystem.imagehandling.data.AdditionalFile;
import org.lsst.ccs.subsystem.imagehandling.data.HeaderServiceEnabled;
import org.lsst.ccs.subsystem.imagehandling.data.ImageHeaderData;
import org.lsst.ccs.subsystem.imagehandling.data.ImageHeaderData.Header;
import org.lsst.ccs.subsystem.imagehandling.data.ImageHeaderKeywords;
import org.lsst.ccs.subsystem.imagehandling.data.MeasuredShutterTime;
import org.lsst.ccs.subsystem.imagehandling.data.TriggerSimulatedImage;
import org.lsst.ccs.utilities.location.Location;
import org.lsst.ccs.utilities.location.LocationSet;

/**
 * A class that listens to messages coming from the focal-plane subsystem.
 *
 * @author The LSST CCS Team
 */
class BusMessageHandler implements HasLifecycle {

    @SuppressWarnings("VolatileArrayField")
    private volatile String[] metaDataRegisters;
    private FocalPlaneConfigurationAndDataListener focalPlaneListener;
    private OCSBridgeDataListener ocsBridgeDataListener;
    private MCMDataListener mcmDataListener;
    private AdditionalFileListener additionalFileListener;

    private static final Logger LOG = Logger.getLogger(BusMessageHandler.class.getName());

    private final Predicate<BusMessage<? extends Serializable, ?>> focalPlanePredicate = (bm) -> isFocalPlaneAgent(bm.getOriginAgentInfo());
    private final Predicate<BusMessage<? extends Serializable, ?>> ocsBridgePredicate = (bm) -> isOCSBridgeAgent(bm.getOriginAgentInfo());
    private final Predicate<BusMessage<? extends Serializable, ?>> mcmPredicate = (bm) -> isMCMAgent(bm.getOriginAgentInfo());

    private final Predicate<BusMessage<? extends Serializable, ?>> headerKeywordsPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> ImageHeaderKeywords.IMAGE_HEADER_KEYWORDS.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> headerDataPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> ImageHeaderData.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> simulatedTriggerPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> TriggerSimulatedImage.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> headerServiceEnabledPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> HeaderServiceEnabled.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> additionalFilePredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> AdditionalFile.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));
   
    private final Predicate<BusMessage<? extends Serializable, ?>> measuredShutterTimePredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> MeasuredShutterTime.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));
    
    private final Predicate<BusMessage<? extends Serializable, ?>> configurationDataPredicate
            = BusMessageFilterFactory.messageClass(StatusConfigurationInfo.class);

    @LookupField(strategy = LookupField.Strategy.TOP)
    private Agent agent;

    @LookupField(strategy = LookupField.Strategy.TREE)
    private ImageHandlingClient imageHandlingClient;

    @Override
    public void postStart() {
        // Called in postStart so the listener is registered (and thus called) after the FitsService (hopefully)
        focalPlaneListener = new FocalPlaneConfigurationAndDataListener();
        agent.getMessagingAccess().addStatusMessageListener(focalPlaneListener, focalPlanePredicate.and(configurationDataPredicate.or(headerKeywordsPredicate).or(simulatedTriggerPredicate)));

        ocsBridgeDataListener = new OCSBridgeDataListener();
        agent.getMessagingAccess().addStatusMessageListener(ocsBridgeDataListener, ocsBridgePredicate.and(headerDataPredicate.or(headerServiceEnabledPredicate)));

        mcmDataListener = new MCMDataListener();
        agent.getMessagingAccess().addStatusMessageListener(mcmDataListener, mcmPredicate.and(measuredShutterTimePredicate));
        
        if (imageHandlingClient.canHandleAdditionalFiles()) {
            additionalFileListener = new AdditionalFileListener();
            agent.getMessagingAccess().addStatusMessageListener(additionalFileListener, additionalFilePredicate);
        }

    }

    @Override
    public void shutdown() {
        agent.getMessagingAccess().removeStatusMessageListener(focalPlaneListener);
        agent.getMessagingAccess().removeStatusMessageListener(ocsBridgeDataListener);
        agent.getMessagingAccess().removeStatusMessageListener(additionalFileListener);
    }

    public String[] getMetaDataRegisters() {
        return metaDataRegisters;
    }

    private static boolean isFocalPlaneAgent(AgentInfo agentInfo) {
        return AgentCategory.FOCAL_PLANE.name().equals(agentInfo.getAgentProperty(AgentCategory.AGENT_CATEGORY_PROPERTY));
    }

    private static boolean isOCSBridgeAgent(AgentInfo agentInfo) {
        return AgentType.OCS_BRIDGE.equals(agentInfo.getType());
    }

    private static boolean isMCMAgent(AgentInfo agentInfo) {
        return AgentType.MCM.equals(agentInfo.getType());
    }

    private class FocalPlaneConfigurationAndDataListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {
            if (msg instanceof StatusConfigurationInfo) {
                ConfigurationInfo ci = ((StatusConfigurationInfo) msg).getConfigurationInfo();
                for (ConfigurationParameterInfo cpi : ci.getAllParameterInfo()) {
                    if (cpi.getComponentName().equals("focalPlaneConfig")) {
                        if (cpi.getParameterName().equals("metaDataRegisters")) {
                            metaDataRegisters = (String[]) cpi.getCurrentValueObject();
                            LOG.log(Level.FINEST, "Updating metaDataRegister to {0}", new Object[]{Arrays.asList(metaDataRegisters)});
                        }
                    }
                }
            }

            if (msg instanceof StatusSubsystemData) {
                StatusSubsystemData ssd = (StatusSubsystemData) msg;
                if (headerKeywordsPredicate.test(ssd)) {
                    ImageHeaderKeywords keywords = (ImageHeaderKeywords) ssd.getObject().getValue();
                    Map<String, Serializable> headerMap = keywords.getKeywords();
                    LOG.log(Level.INFO, () -> String.format("Received image header keywords %s", keywords));
                    // LSSTCCSRAFTS-678 We collect here the daq metadata and forward it to the imageHandlingClient
                    imageHandlingClient.addMetaData(keywords.getImageName(), keywords.getKeywords());
                    if (headerMap.containsKey("darkTime")) {
                        imageHandlingClient.darkTimeArrived(keywords.getImageName());
                    } else if (headerMap.containsKey("RunNumber")) {
                        Object runNumber = headerMap.get("RunNumber");
                        imageHandlingClient.runNumberArrived(runNumber == null ? null : runNumber.toString());
                    }
                } else if (simulatedTriggerPredicate.test(ssd)) {
                    TriggerSimulatedImage tsi = (TriggerSimulatedImage) ssd.getObject().getValue();
                    LOG.log(Level.INFO, "Received {0}", tsi);
                    try {
                        // Hardwird for AuxTel for now
                        Location location = Location.of("R00/RebW");
                        ImageMetaData meta = new ImageMetaData(tsi.getTriggeredImageName(), tsi.getTriggeredDAQFolder(), tsi.getAnnotation(), tsi.getOpCode(), new LocationSet(Collections.singleton(location)));
                        // ToDo, include emulated DAQ folder in this? Move the logic into the playlist implementation?
                        Path metaPath = Path.of(tsi.getEmulatedDataFolder()).resolve(tsi.getEmulatedDAQFolder()).resolve(tsi.getEmulatedImageName() + "_R00_RebW.meta");
                        if (!Files.exists(metaPath)) {
                            throw new IllegalArgumentException("Path " + metaPath + " does not exist");
                        }
                        Path rawData = Path.of(tsi.getEmulatedDataFolder()).resolve(tsi.getEmulatedDAQFolder()).resolve(tsi.getEmulatedImageName() + "_R00_RebW.raw");
                        if (!Files.exists(rawData)) {
                            throw new IllegalArgumentException("Path " + rawData + " does not exist");
                        }
                        String line = Files.newBufferedReader(metaPath).readLine();
                        int[] registerList = Arrays.stream(line.substring(1, line.length() - 1).split(",")).map(String::trim).mapToInt(Integer::parseInt).toArray();
                        imageHandlingClient.simulateTrigger(location, meta, registerList, rawData);
                    } catch (IOException | DAQException x) {
                        LOG.log(Level.SEVERE, "Error triggering simulated image", x);
                    }

                }

            }
        }
    }

    private class OCSBridgeDataListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {

            if (msg instanceof StatusSubsystemData) {
                StatusSubsystemData ssd = (StatusSubsystemData) msg;
                if (headerDataPredicate.test(ssd)) {
                    ImageHeaderData data = (ImageHeaderData) ssd.getObject().getValue();
                    // Figure out what image these headers correspond to, and pass to the appropriate client
                    // Unfortunately header service does not provide the OBSID so we need to put the pieces together
                    StringBuilder imageName = new StringBuilder("aa_b_cccccccc_dddddd");
                    for (Header header : data.getHeaders()) {
                        if (header.getKeyword() == null) {
                            continue;
                        }
                        final String value = header.getValue();
                        switch (header.getKeyword()) {
                            case "CAMCODE":
                                imageName.replace(0, 2, value);
                                break;
                            case "CONTRLLR":
                                imageName.replace(3, 4, value);
                                break;
                            case "DAYOBS":
                                imageName.replace(5, 13, value);
                                break;
                            case "SEQNUM":
                                imageName.replace(14, 20, String.format("%06d", Integer.parseInt(value)));
                                break;
                        }
                    }
                    LOG.log(Level.INFO, "Received image headers for {0}: {1}", new Object[]{imageName, data.getHeaders()});
                    imageHandlingClient.headerDataArrived(imageName.toString(), data.getHeaders());
                } else if (headerServiceEnabledPredicate.test(ssd)) {
                    HeaderServiceEnabled headerServiceEnabled = (HeaderServiceEnabled) ssd.getObject().getValue();
                    LOG.log(Level.INFO, "Received headerServiceEnabled: {0}", headerServiceEnabled.isEnabled());
                    imageHandlingClient.setHeaderServiceEnabled(headerServiceEnabled.isEnabled());
                }
            }
        }
    }

    private class MCMDataListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {

            if (msg instanceof StatusSubsystemData) {
                StatusSubsystemData ssd = (StatusSubsystemData) msg;
                if (measuredShutterTimePredicate.test(ssd)) {
                    MeasuredShutterTime shutterTime = (MeasuredShutterTime) ssd.getObject().getValue();
                    LOG.log(Level.INFO, "Received ", shutterTime);
                    // pass it to imageHandlingclient 
                    imageHandlingClient.handleMeasuredShutterTime(shutterTime);
                } 
            }
        }
    }

    private class AdditionalFileListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {
            if (msg instanceof StatusSubsystemData) {
                StatusSubsystemData ssd = (StatusSubsystemData) msg;
                if (additionalFilePredicate.test(ssd)) {
                    AdditionalFile additionalFile = (AdditionalFile) ssd.getObject().getValue();
                    imageHandlingClient.handleAdditionFile(additionalFile);
                }
            }
        }
    }
}

==> org-lsst-ccs-subsystem-image-handling-main-1.1.10-sources.jar: org/lsst/ccs/subsystem/imagehandling/BusMessageHandler.java <==
package org.lsst.ccs.subsystem.imagehandling;

import java.io.IOException;
import java.io.Serializable;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.lsst.ccs.Agent;
import org.lsst.ccs.bus.data.AgentCategory;
import org.lsst.ccs.bus.data.AgentInfo;
import org.lsst.ccs.bus.data.AgentInfo.AgentType;
import org.lsst.ccs.bus.data.ConfigurationInfo;
import org.lsst.ccs.bus.data.ConfigurationParameterInfo;
import org.lsst.ccs.bus.messages.BusMessage;
import org.lsst.ccs.bus.messages.StatusConfigurationInfo;
import org.lsst.ccs.bus.messages.StatusMessage;
import org.lsst.ccs.bus.messages.StatusSubsystemData;
import org.lsst.ccs.commons.annotations.LookupField;
import org.lsst.ccs.daq.ims.DAQException;
import org.lsst.ccs.daq.ims.ImageMetaData;
import org.lsst.ccs.framework.HasLifecycle;
import org.lsst.ccs.imagenaming.ImageName;
import org.lsst.ccs.messaging.BusMessageFilterFactory;
import org.lsst.ccs.messaging.StatusMessageListener;
import org.lsst.ccs.subsystem.imagehandling.data.AdditionalFile;
import org.lsst.ccs.subsystem.imagehandling.data.HeaderServiceEnabled;
import org.lsst.ccs.subsystem.imagehandling.data.ImageHeaderData;
import org.lsst.ccs.subsystem.imagehandling.data.ImageHeaderData.Header;
import org.lsst.ccs.subsystem.imagehandling.data.ImageHeaderKeywords;
import org.lsst.ccs.subsystem.imagehandling.data.MeasuredShutterTime;
import org.lsst.ccs.subsystem.imagehandling.data.TriggerSimulatedImage;
import org.lsst.ccs.utilities.location.Location;
import org.lsst.ccs.utilities.location.LocationSet;

/**
 * A class that listens to messages coming from the focal-plane subsystem.
 *
 * @author The LSST CCS Team
 */
class BusMessageHandler implements HasLifecycle {

    @SuppressWarnings("VolatileArrayField")
    private volatile String[] metaDataRegisters;
    private FocalPlaneConfigurationAndDataListener focalPlaneListener;
    private OCSBridgeDataListener ocsBridgeDataListener;
    private MCMDataListener mcmDataListener;
    private AdditionalFileListener additionalFileListener;

    private static final Logger LOG = Logger.getLogger(BusMessageHandler.class.getName());

    private final Predicate<BusMessage<? extends Serializable, ?>> focalPlanePredicate = (bm) -> isFocalPlaneAgent(bm.getOriginAgentInfo());
    private final Predicate<BusMessage<? extends Serializable, ?>> ocsBridgePredicate = (bm) -> isOCSBridgeAgent(bm.getOriginAgentInfo());
    private final Predicate<BusMessage<? extends Serializable, ?>> mcmPredicate = (bm) -> isMCMAgent(bm.getOriginAgentInfo());

    private final Predicate<BusMessage<? extends Serializable, ?>> headerKeywordsPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> ImageHeaderKeywords.IMAGE_HEADER_KEYWORDS.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> headerDataPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> ImageHeaderData.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> simulatedTriggerPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> TriggerSimulatedImage.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> headerServiceEnabledPredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> HeaderServiceEnabled.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));

    private final Predicate<BusMessage<? extends Serializable, ?>> additionalFilePredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> AdditionalFile.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));
   
    private final Predicate<BusMessage<? extends Serializable, ?>> measuredShutterTimePredicate
            = BusMessageFilterFactory.messageClass(StatusSubsystemData.class).and((bm) -> MeasuredShutterTime.EVENT_KEY.equals(((StatusSubsystemData) bm).getDataKey()));
    
    private final Predicate<BusMessage<? extends Serializable, ?>> configurationDataPredicate
            = BusMessageFilterFactory.messageClass(StatusConfigurationInfo.class);

    @LookupField(strategy = LookupField.Strategy.TOP)
    private Agent agent;

    @LookupField(strategy = LookupField.Strategy.TREE)
    private ImageHandlingClient imageHandlingClient;

    @Override
    public void postStart() {
        // Called in postStart so the listener is registered (and thus called) after the FitsService (hopefully)
        focalPlaneListener = new FocalPlaneConfigurationAndDataListener();
        agent.getMessagingAccess().addStatusMessageListener(focalPlaneListener, focalPlanePredicate.and(configurationDataPredicate.or(headerKeywordsPredicate).or(simulatedTriggerPredicate)));

        ocsBridgeDataListener = new OCSBridgeDataListener();
        agent.getMessagingAccess().addStatusMessageListener(ocsBridgeDataListener, ocsBridgePredicate.and(headerDataPredicate.or(headerServiceEnabledPredicate)));

        mcmDataListener = new MCMDataListener();
        agent.getMessagingAccess().addStatusMessageListener(mcmDataListener, mcmPredicate.and(measuredShutterTimePredicate));
        
        if (imageHandlingClient.canHandleAdditionalFiles()) {
            additionalFileListener = new AdditionalFileListener();
            agent.getMessagingAccess().addStatusMessageListener(additionalFileListener, additionalFilePredicate);
        }

    }

    @Override
    public void shutdown() {
        agent.getMessagingAccess().removeStatusMessageListener(focalPlaneListener);
        agent.getMessagingAccess().removeStatusMessageListener(ocsBridgeDataListener);
        agent.getMessagingAccess().removeStatusMessageListener(additionalFileListener);
    }

    public String[] getMetaDataRegisters() {
        return metaDataRegisters;
    }

    private static boolean isFocalPlaneAgent(AgentInfo agentInfo) {
        return AgentCategory.FOCAL_PLANE.name().equals(agentInfo.getAgentProperty(AgentCategory.AGENT_CATEGORY_PROPERTY));
    }

    private static boolean isOCSBridgeAgent(AgentInfo agentInfo) {
        return AgentType.OCS_BRIDGE.equals(agentInfo.getType());
    }

    private static boolean isMCMAgent(AgentInfo agentInfo) {
        return AgentType.MCM.equals(agentInfo.getType());
    }

    private class FocalPlaneConfigurationAndDataListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {
            if (msg instanceof StatusConfigurationInfo statusConfigurationInfo) {
                ConfigurationInfo ci = statusConfigurationInfo.getConfigurationInfo();
                for (ConfigurationParameterInfo cpi : ci.getAllParameterInfo()) {
                    if (cpi.getComponentName().equals("focalPlaneConfig")) {
                        if (cpi.getParameterName().equals("metaDataRegisters")) {
                            metaDataRegisters = (String[]) cpi.getCurrentValueObject();
                            LOG.log(Level.FINEST, "Updating metaDataRegister to {0}", new Object[]{Arrays.asList(metaDataRegisters)});
                        }
                    }
                }
            }

            if (msg instanceof StatusSubsystemData ssd) {
                if (headerKeywordsPredicate.test(ssd)) {
                    ImageHeaderKeywords keywords = (ImageHeaderKeywords) ssd.getObject().getValue();
                    Map<String, Serializable> headerMap = keywords.getKeywords();
                    LOG.log(Level.INFO, () -> String.format("Received image header keywords %s", keywords));
                    // LSSTCCSRAFTS-678 We collect here the daq metadata and forward it to the imageHandlingClient
                    imageHandlingClient.addMetaData(keywords.getImageName(), keywords.getKeywords());
                    if (headerMap.containsKey("darkTime")) {
                        imageHandlingClient.darkTimeArrived(keywords.getImageName());
                    } else if (headerMap.containsKey("RunNumber")) {
                        Object runNumber = headerMap.get("RunNumber");
                        imageHandlingClient.runNumberArrived(runNumber == null ? null : runNumber.toString());
                    }
                } else if (simulatedTriggerPredicate.test(ssd)) {
                    TriggerSimulatedImage tsi = (TriggerSimulatedImage) ssd.getObject().getValue();
                    LOG.log(Level.INFO, "Received {0}", tsi);
                    try {
                        // Hardwird for AuxTel for now
                        Location location = Location.of("R00/RebW");
                        ImageMetaData meta = new ImageMetaData(tsi.getTriggeredImageName(), tsi.getTriggeredDAQFolder(), tsi.getAnnotation(), tsi.getOpCode(), new LocationSet(Collections.singleton(location)));
                        // ToDo, include emulated DAQ folder in this? Move the logic into the playlist implementation?
                        Path metaPath = Path.of(tsi.getEmulatedDataFolder()).resolve(tsi.getEmulatedDAQFolder()).resolve(tsi.getEmulatedImageName() + "_R00_RebW.meta");
                        if (!Files.exists(metaPath)) {
                            throw new IllegalArgumentException("Path " + metaPath + " does not exist");
                        }
                        Path rawData = Path.of(tsi.getEmulatedDataFolder()).resolve(tsi.getEmulatedDAQFolder()).resolve(tsi.getEmulatedImageName() + "_R00_RebW.raw");
                        if (!Files.exists(rawData)) {
                            throw new IllegalArgumentException("Path " + rawData + " does not exist");
                        }
                        String line = Files.newBufferedReader(metaPath).readLine();
                        int[] registerList = Arrays.stream(line.substring(1, line.length() - 1).split(",")).map(String::trim).mapToInt(Integer::parseInt).toArray();
                        imageHandlingClient.simulateTrigger(location, meta, registerList, rawData);
                    } catch (IOException | DAQException x) {
                        LOG.log(Level.SEVERE, "Error triggering simulated image", x);
                    }

                }

            }
        }
    }

    private class OCSBridgeDataListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {

            if (msg instanceof StatusSubsystemData ssd) {
                if (headerDataPredicate.test(ssd)) {
                    ImageHeaderData data = (ImageHeaderData) ssd.getObject().getValue();
                    // Figure out what image these headers correspond to, and pass to the appropriate client
                    StringBuilder imageName = new StringBuilder("aa_b_cccccccc_dddddd");
                    ImageName obsid = null;
                    for (Header header : data.getHeaders()) {
                        if (header.getKeyword() == null) {
                            continue;
                        }
                        final String value = header.getValue();
                        switch (header.getKeyword()) {
                            case "CAMCODE" -> imageName.replace(0, 2, value);
                            case "CONTRLLR" -> imageName.replace(3, 4, value);
                            case "DAYOBS" -> imageName.replace(5, 13, value);
                            case "SEQNUM" -> imageName.replace(14, 20, String.format("%06d", Integer.valueOf(value)));
                            case "OBSID" -> obsid = new ImageName(value);
                        }
                    }
                    // Historic, OBSID should always be present now.
                    if (obsid == null) obsid = new ImageName(imageName.toString());
                    LOG.log(Level.INFO, "Received image headers for {0}: {1}", new Object[]{obsid, data.getHeaders()});
                    imageHandlingClient.headerDataArrived(obsid, data.getHeaders());
                } else if (headerServiceEnabledPredicate.test(ssd)) {
                    HeaderServiceEnabled headerServiceEnabled = (HeaderServiceEnabled) ssd.getObject().getValue();
                    LOG.log(Level.INFO, "Received headerServiceEnabled: {0}", headerServiceEnabled.isEnabled());
                    imageHandlingClient.setHeaderServiceEnabled(headerServiceEnabled.isEnabled());
                }
            }
        }
    }

    private class MCMDataListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {

            if (msg instanceof StatusSubsystemData ssd) {
                if (measuredShutterTimePredicate.test(ssd)) {
                    MeasuredShutterTime shutterTime = (MeasuredShutterTime) ssd.getObject().getValue();
                    LOG.log(Level.INFO, "Received ", shutterTime);
                    // pass it to imageHandlingclient 
                    imageHandlingClient.handleMeasuredShutterTime(shutterTime);
                } 
            }
        }
    }

    private class AdditionalFileListener implements StatusMessageListener {

        @Override
        public void onStatusMessage(StatusMessage msg) {
            if (msg instanceof StatusSubsystemData ssd) {
                if (additionalFilePredicate.test(ssd)) {
                    AdditionalFile additionalFile = (AdditionalFile) ssd.getObject().getValue();
                    imageHandlingClient.handleAdditionFile(additionalFile);
                }
            }
        }
    }
}
